#include <LiquidCrystal.h>
#include <EEPROM.h>

// v1.1, 06/12/2021: 16-bit EEPROM write/read issue fixed

LiquidCrystal lcd(8, 9, 4, 5, 6, 7);

int lcd_key     = 0;
int adc_key_in  = 0;
int lastkey     = 5;                // initialise to no button pressed
boolean entry_mode  = false;        // entry_mode false = line select, true = value select

int LEDpin = 13;
int tValve = 2;
int tFlash = 3;
int tCamera = 11;

#define btnRIGHT  0
#define btnUP     1
#define btnDOWN   2
#define btnLEFT   3
#define btnSELECT 4
#define btnNONE   5

int backLightControl = 10;          // Output to control back light on LCD
int MessageLoop;

int FlashDelay;
int DropCount;
int DropNumber;
int DropSize[3];
int DropDelay[2];
String MessageArray[] = {"No of Drops  ","Drop 1 Size  ","Gap 1 Length ","Drop 2 Size  ","Gap 2 Length ","Drop 3 Size  ","Flash Delay  ", "Fire         "};
int ParamValues[8] = {1,10,100,20,200,30,300,0};
int ParamMax[8] = {3,99,999,99,999,99,999,1};

int msg_point = 0;
int msg_point2 = 1;
int x_pos = 13;                                                                         // cursor position
int pos;
int fireoption = 0;
 
int valstep = 1;                                                                        // size of step to increment/decrement by

int read_keypad_buttons(){                                                              // read the buttons

    adc_key_in = analogRead(0);                                                         // read the value from the sensor
    delay(200);                                                                         // delay to avoid multiple readings per key press
    if (adc_key_in > 1000) return btnNONE;

    // For V1.0 use these thresholds
    if (adc_key_in < 50) return btnRIGHT;  
    if (adc_key_in < 195) return btnUP; 
    if (adc_key_in < 380) return btnDOWN; 
    if (adc_key_in < 555) return btnLEFT; 
    if (adc_key_in < 790) return btnSELECT;  
    return btnNONE;
}

void setup(){
  pinMode(backLightControl,OUTPUT);
  pinMode(tValve,OUTPUT);                                                               // valve control pin as output
  pinMode(tFlash,OUTPUT);                                                               // set flash trigger pin to be output
  pinMode(tCamera,OUTPUT);                                                              // set camera trigger pin as output

  digitalWrite(tValve,LOW);                                                             // them set them all low to start
  digitalWrite(tFlash,LOW);
  digitalWrite(tCamera,LOW);
  Serial.begin(9600);

  for (int i=0; i<=6; i++) {
      EEPROM.get((2 * i), ParamValues[i]);                                              // read last settings used from EEPROM
  }
  ParamValues[0] = ParamValues[0] & 0b0000000011111111;                                 // mask top byte for no of drops
  ParamValues[7] = ParamValues[7] & 0b0000000011111111;                                 // mask top byte for fire
  
  lcd.begin(16,2);
  lcd.setCursor(0,0);
  lcd.print("Drop Controller");
  lcd.setCursor(0,1);
  lcd.print("photobuilds.com");
  digitalWrite(backLightControl,HIGH);                                              // turn display backlight on
  delay(2000);
  lcd.setCursor(0,0);
  lcd.print("                ");                                                        // clear the top display line
  lcd.setCursor(0,1);
  lcd.print("                ");                                                        // clear the bottom display line
  lcd.setCursor(0,0);
  lcd.print(MessageArray[0]);
  lcd.setCursor(15,0);
  lcd.print(ParamValues[0]);
  lcd.setCursor(0,1);
  lcd.print(MessageArray[1]);
  if (ParamValues[1] < 10) lcd.setCursor(15,1);
  else lcd.setCursor(14,1);
  lcd.print(ParamValues[1]);
}

void loop() {

  digitalWrite(backLightControl,HIGH);                                                    // turn display backlight on

  lcd_key = read_keypad_buttons();                                                        // read if the last keypress has been used
  if (lcd_key != lastkey) {
    lastkey = lcd_key;  
  
    Serial.print("Entry mode ");
    Serial.println(entry_mode);
    switch (lcd_key){                                                                     // depending on which button was pushed, we perform an action
      case btnDOWN:{                                                                      //  push button "RIGHT" and show the word on the screen
        if (entry_mode==false) {
          msg_point = msg_point + 1;
          if (msg_point==8) {
            msg_point =0; 
          }
          break;
        }
        else {
          ParamValues[msg_point] = ParamValues[msg_point] - valstep;
          if (ParamValues[msg_point]<1) ParamValues[msg_point]=1;
          lcd.setCursor(13,0);
          lcd.print("   ");                                                               // clear the value area of top display line
          if (ParamValues[msg_point] < 10) lcd.setCursor(15,0);
          else if (ParamValues[msg_point] < 100) lcd.setCursor(14,0);
          else lcd.setCursor(13,0);
          lcd.print(ParamValues[msg_point]);
          lcd.setCursor(x_pos,0);
          lcd.blink();
          break;
        }
      }
      case btnUP:{
        if (entry_mode==false) {
          msg_point = msg_point-1;
          if (msg_point<0) {
            msg_point =7; 
          }
          break;
        }
        else {
          ParamValues[msg_point] = ParamValues[msg_point] + valstep;
          if (ParamValues[msg_point] > ParamMax[msg_point]) {                             // trap for maximum values
            ParamValues[msg_point] = ParamMax[msg_point];                                 // if greater, limit to maximum value
          }
          lcd.setCursor(13,0);
          lcd.print("   ");                                                               // clear the value area of top display line
          if (ParamValues[msg_point] < 10) lcd.setCursor(15,0);
          else if (ParamValues[msg_point] < 100) lcd.setCursor(14,0);
          else lcd.setCursor(13,0);

          lcd.print(ParamValues[msg_point]);
          lcd.setCursor(x_pos,0);
          lcd.blink();
          break; 
        }
      }    
      case btnRIGHT:{
        if (entry_mode ==true) {
          x_pos = x_pos + 1;
          if (x_pos == 16) x_pos = 15;
          if (x_pos==13) valstep = 100;
          else if (x_pos==14) valstep = 10;
          else valstep = 1;
        
          lcd.setCursor(x_pos,0);
          lcd.blink();
         }
        break;
      }
      case btnLEFT:{
        if (entry_mode==true) {
          x_pos = x_pos - 1;
          if (msg_point==1 || msg_point==3 || msg_point==5) {                             // only allowing two digits for drop sizes
            if (x_pos < 14) x_pos=14;
          }
          else if (msg_point==2 || msg_point==4 || msg_point==6) {                        // allowing three digits for gaps and flash delay
            if (x_pos < 13) x_pos=13;
          }
          else {
            if (x_pos < 15) x_pos=15;                                                     // and a single digit for number of drops
          }
          if (x_pos==13) valstep = 100;
          else if (x_pos==14) valstep = 10;
          else valstep = 1;

          lcd.setCursor(x_pos,0);
          lcd.blink();
          delay(100);
        }
        break;
      }
      case btnSELECT:{
        entry_mode = !entry_mode;
        if (entry_mode==true) {
          if (msg_point < 7) {
            lcd.setCursor(0,0);
            lcd.print(MessageArray[msg_point]);
            if (ParamValues[msg_point] < 10) lcd.setCursor(15,0);
            else if (ParamValues[msg_point] < 100) lcd.setCursor(14,0);
            else lcd.setCursor(13,0);
            lcd.print(ParamValues[msg_point]);
            lcd.setCursor(0,1);
            lcd.print(MessageArray[msg_point2]);
            if (ParamValues[msg_point2] < 10) lcd.setCursor(15,1);
            else if (ParamValues[msg_point2] < 100) lcd.setCursor(14,1);
            else lcd.setCursor(13,1);
            lcd.print(ParamValues[msg_point2]);

            valstep = 1;                                                                    // set step to 1 and column to units for new selection
            x_pos = 15;
            lcd.setCursor(x_pos,0);
            lcd.blink();
          }
          else {
            for (int i=0; i<=6; i++) {
              EEPROM.put((2 * i), ParamValues[i]);                                          // write the current values to EEPROM
            }
            TriggerValve();                                                                 // trigger the valve
            entry_mode = !entry_mode;                                                       // then come straight back out of entry mode
          }
        }
        else {
          lcd.noBlink();
        }
        break;
      }
      case btnNONE:{
        break;                                                                            // do nothing
      }
    }
    if (entry_mode==false) {
        lcd.setCursor(0,0);
        lcd.print("                ");                                                    // clear the top display line
        lcd.setCursor(0,0);
        lcd.print(MessageArray[msg_point]);
        if (ParamValues[msg_point] < 10) lcd.setCursor(15,0);
        else if (ParamValues[msg_point] < 100) lcd.setCursor(14,0);
        else lcd.setCursor(13,0);
        lcd.print(ParamValues[msg_point]);
        lcd.setCursor(0,1);
        if (msg_point==7) {
          msg_point2 =0; 
        }
        else {
          msg_point2 = msg_point+1;
        }
        lcd.print("                ");                                                    // clear the bottom display line
        lcd.setCursor(0,1);
        lcd.print(MessageArray[msg_point2]);
        if (ParamValues[msg_point2] < 10) lcd.setCursor(15,1);
        else if (ParamValues[msg_point2] < 100) lcd.setCursor(14,1);
        else lcd.setCursor(13,1);
        lcd.print(ParamValues[msg_point2]);
      }
    }
  }


void TriggerValve() {

int stepup = 0;

  digitalWrite(backLightControl,LOW);
  delay(250);                                                                             // wait quarter second after pressing the button
  digitalWrite(tCamera,HIGH);                                                             // trigger camera
  delay(100);                                                                             // give the camera time to settle after triggering
  for ( int dropcnt=1; dropcnt<=ParamValues[0]; dropcnt++) {
    digitalWrite(LEDpin,HIGH);
    digitalWrite(tValve,HIGH);                                                            // open valve
    delay(ParamValues[dropcnt+stepup]);
    Serial.print("Drop: ");
    Serial.print(dropcnt);
    Serial.print(" Size: ");
    Serial.println(ParamValues[dropcnt+stepup]);
    digitalWrite(tValve,LOW);                                                             // close valve
    digitalWrite(LEDpin,LOW);
    if (dropcnt < ParamValues[0]) {
      delay(ParamValues[dropcnt+stepup+1]);                                               // delay between drops
      Serial.print("Delay: ");
      Serial.println(ParamValues[dropcnt+stepup+1]);
      stepup=stepup+1;
    }
  }
  delay(ParamValues[6]);                                                                  // delay before flash trigger 
  Serial.print("Flash delay ");
  Serial.println(ParamValues[6]);

  digitalWrite(tFlash,HIGH);                                                              // trigger flash
  delay(15);                                                                              // hold high long enough to actually trigger the flash
  digitalWrite(tFlash,LOW);
  digitalWrite(tCamera,LOW);                                                            
  digitalWrite(backLightControl,HIGH);
}
